from types import SimpleNamespace

from kabomu.misc_utils_internal import parse_int_32

# Request environment variable for local server endpoint.
ENV_KEY_LOCAL_PEER_ENDPOINT = "kabomu.local_peer_endpoint"

# Request environment variable for remote client endpoint.
ENV_KEY_REMOTE_PEER_ENDPOINT = "kabomu.remote_peer_endpoint"

# Request environment variable for the transport instance from
# which a request was received.
ENV_KEY_TRANSPORT_INSTANCE = "kabomu.transport"

# Request environment variable for the connection from which a
# request was received.
ENV_KEY_CONNECTION = "kabomu.connection"

METHOD_CONNECT = "CONNECT"
METHOD_DELETE = "DELETE"
METHOD_GET = "GET"
METHOD_HEAD = "HEAD"
METHOD_OPTIONS = "OPTIONS"
METHOD_PATCH = "PATCH"
METHOD_POST = "POST"
METHOD_PUT = "PUT"
METHOD_TRACE = "TRACE"

# 200 OK
STATUS_CODE_OK = 200

# 400 Bad Request
STATUS_CODE_CLIENT_ERROR_BAD_REQUEST = 400

# 401 Unauthorized
STATUS_CODE_CLIENT_ERROR_UNAUTHORIZED = 401

# 403 Forbidden
STATUS_CODE_CLIENT_ERROR_FORBIDDEN = 403

# 404 Not Found
STATUS_CODE_CLIENT_ERROR_NOT_FOUND = 404

# 405 Method Not Allowed
STATUS_CODE_CLIENT_ERROR_METHOD_NOT_ALLOWED = 405

# 413 Payload Too Large
STATUS_CODE_CLIENT_ERROR_PAYLOAD_TOO_LARGE = 413

# 414 URI Too Long
STATUS_CODE_CLIENT_ERROR_URI_TOO_LONG = 414

# 415 Unsupported Media Type
STATUS_CODE_CLIENT_ERROR_UNSUPPORTED_MEDIA_TYPE = 415

# 422 Unprocessable Entity
STATUS_CODE_CLIENT_ERROR_UNPROCESSABLE_ENTITY = 422

# 429 Too Many Requests
STATUS_CODE_CLIENT_ERROR_TOO_MANY_REQUESTS = 429

# 500 Internal Server Error
STATUS_CODE_SERVER_ERROR = 500

# The default value of maximum size of headers in a request or response.
DEFAULT_MAX_HEADERS_SIZE = 8_192

def _get_optional_attr(instance, *args):
    for n in args:
        if instance == None or not hasattr(instance, n):
            return None
        instance = getattr(instance, n)
    return instance

def _bind_method(proc, instance):
    return proc.__get__(instance, instance.__class__)

def merge_processing_options(preferred, fallback):
    if preferred is None or fallback is None:
        return preferred if preferred is not None else fallback
    merged_options = SimpleNamespace()
    merged_options.timeout_millis =\
        _determine_effective_non_zero_integer_option(
            _get_optional_attr(preferred, "timeout_millis"),
            _get_optional_attr(fallback, "timeout_millis"),
        0)
    merged_options.extra_connectivity_params =\
        _determine_effective_options(
            _get_optional_attr(preferred, "extra_connectivity_params"),
            _get_optional_attr(fallback, "extra_connectivity_params"))
    merged_options.max_headers_size =\
        _determine_effective_positive_integer_option(
            _get_optional_attr(preferred, "max_headers_size"),
            _get_optional_attr(fallback, "max_headers_size"),
        0)
    merged_options.max_response_body_size =\
        _determine_effective_non_zero_integer_option(
            _get_optional_attr(preferred, "max_response_body_size"),
            _get_optional_attr(fallback, "max_response_body_size"),
        0)
    return merged_options

def _determine_effective_non_zero_integer_option(
        preferred, fallback1, default_value):
    if preferred:
        return parse_int_32(preferred)
    if fallback1:
        return parse_int_32(fallback1)
    return parse_int_32(default_value)

def _determine_effective_positive_integer_option(
        preferred, fallback1, default_value):
    if preferred:
        effective_value = parse_int_32(preferred)
        if effective_value > 0:
            return effective_value
    if fallback1:
        effective_value = parse_int_32(fallback1)
        if effective_value > 0:
            return effective_value
    return parse_int_32(default_value)

def _determine_effective_options(
        preferred, fallback):
    dest = {}
    # since we want preferred options to overwrite fallback options,
    # set fallback options first.
    if fallback:
        dest.update(**fallback)
    if preferred:
        dest.update(**preferred)
    return dest



import abc


class IQuasiHttpAltTransport(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def request_serializer(self, connection, request):
        raise NotImplementedError()

    @abc.abstractmethod
    async def response_serializer(self, connection, response):
        raise NotImplementedError()

    @abc.abstractmethod
    async def request_deserializer(self, connection):
        raise NotImplementedError()

    @abc.abstractmethod
    async def response_deserializer(self, connection):
        raise NotImplementedError()


class IQuasiHttpClientTransport(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def allocate_connection(self, remote_endpoint, send_options):
        raise NotImplementedError()
    
    @abc.abstractmethod
    async def establish_connection(self, connection):
        raise NotImplementedError()
    
    @abc.abstractmethod
    async def release_connection(self, connection, response):
        raise NotImplementedError()
    
    @abc.abstractmethod
    def get_readable_stream(self, connection):
        raise NotImplementedError()
    
    @abc.abstractmethod
    def get_writable_stream(self, connection):
        raise NotImplementedError()


class IQuasiHttpConnection(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def processing_options(self):
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def environment(self):
        raise NotImplementedError()

    @abc.abstractmethod
    async def schedule_timeout(self, proc):
        raise NotImplementedError()

class QuasiHttpProcessingOptions():
    def __init__(self,
                 extra_connectivity_params=None,
                 timeout_millis=None,
                 max_headers_size=None,
                 max_response_body_size=None):
        self.extra_connectivity_params = extra_connectivity_params
        self.timeout_millis = timeout_millis
        self.max_headers_size = max_headers_size
        self.max_response_body_size = max_response_body_size


class DefaultQuasiHttpRequest():
    def __init__(self,
                 target=None,
                 headers=None,
                 content_length=None,
                 body=None,
                 http_method=None,
                 http_version=None,
                 environment=None):
        self.target = target
        self.headers = headers
        self.content_length = content_length
        self.body = body
        self.http_method = http_method
        self.http_version = http_version
        self.environment = environment

    async def release(self):
        body = self.body
        if body:
            await body.aclose()


class DefaultQuasiHttpResponse():
    def __init__(self,
                 status_code=None,
                 headers=None,
                 content_length=None,
                 body=None,
                 http_status_message=None,
                 http_version=None,
                 environment=None):
        self.status_code = status_code
        self.headers = headers
        self.content_length = content_length
        self.body = body
        self.http_status_message = http_status_message
        self.http_version = http_version
        self.environment = environment

    async def release(self):
        body = self.body
        if body:
            await body.aclose()


class IQuasiHttpServerTransport(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    async def release_connection(self, connection):
        raise NotImplementedError()
    
    @abc.abstractmethod
    def get_readable_stream(self, connection):
        raise NotImplementedError()
    
    @abc.abstractmethod
    def get_writable_stream(self, connection):
        raise NotImplementedError()


class IQuasiHttpApplication(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    async def process_request(self, request):
        raise NotImplementedError()

class DefaultTimeoutResult:
    def __init__(self,
                 response=None,
                 timeout=False,
                 error=None):
        self.response = response
        self.timeout = timeout
        self.error = error



import sys

from kabomu.quasi_http_utils import _get_optional_attr, _bind_method
from kabomu import protocol_utils_internal
from kabomu.errors import MissingDependencyError,\
    QuasiHttpError,\
    QUASI_HTTP_ERROR_REASON_GENERAL

class StandardQuasiHttpClient:
    def __init__(self, transport=None):
        self.transport = transport

    async def send(self, remote_endpoint, request, options=None):
        if not request:
            raise ValueError("request argument is null")
        return await self._send_internal(
            remote_endpoint, request, None, options)

    async def send2(self, remote_endpoint, request_func, options=None):
        if not request_func:
            raise ValueError("request_func argument is null")
        return await self._send_internal(
            remote_endpoint, None, request_func, options)

    async def _send_internal(self,
            remote_endpoint, request, request_func,
            send_options):
        # access fields for use per request call, in order to cooperate with
        # any implementation of field accessors which supports
        # concurrent modifications.
        transport = self.transport

        if not transport:
            raise MissingDependencyError("client transport")

        connection = None
        try:
            connection = await transport.allocate_connection(
                remote_endpoint, send_options)
            if not connection:
                raise QuasiHttpError("no connection")

            async def proc():
                return await _process_send(
                    request, request_func,
                    transport, connection)
            response = await protocol_utils_internal.run_timeout_scheduler(
                connection, True, proc)
            if not response:
                response = await _process_send(
                    request, request_func, transport, connection)
            await _abort(transport, connection, False, response)
            return response
        except:
            #raise
            if connection:
                await _abort(transport, connection, True)
            ex = sys.exc_info()[1]
            if isinstance(ex, QuasiHttpError):
                raise
            abort_error = QuasiHttpError(
                QUASI_HTTP_ERROR_REASON_GENERAL,
                "encountered error during send request processing")
            abort_error.__cause__ = ex
            raise abort_error

async def _process_send(request, request_func, transport, connection):
    # wait for connection to be completely established.
    await transport.establish_connection(connection)

    if not request:
        request = await request_func(
            _get_optional_attr(connection, "environment"))
        if not request:
            raise QuasiHttpError("no request")

    # send entire request first before
    # receiving of response.
    request_serialized = False
    if hasattr(transport, "request_serializer"):
        request_serialized = await transport.request_serializer(
            connection, request)
    if not request_serialized:
        await protocol_utils_internal.write_entity_to_transport(
            False, request, transport.get_writable_stream(connection),
            connection)

    response = None
    if hasattr(transport, "response_deserializer"):
        response = await transport.response_deserializer(connection)
    if not response:
        response = await protocol_utils_internal.read_entity_from_transport(
            True, transport.get_readable_stream(connection),
            connection)
        async def release_func(self):
            await transport.release_connection(connection, None)
        response.release = _bind_method(
            release_func, response)
    return response

async def _abort(transport, connection, error_occured, response=None):
    if error_occured:
        try:
            # don't wait
            await transport.release_connection(
                connection, None) # swallow errors
        except:
            pass # ignore
    else:
        await transport.release_connection(
                connection, response)



from trio.abc import HalfCloseableStream, ReceiveStream, SendStream

from kabomu import io_utils_internal, misc_utils_internal
from kabomu.errors import KabomuIOError, IllegalArgumentError

TAG_FOR_QUASI_HTTP_HEADERS = 0x68647273

TAG_FOR_QUASI_HTTP_BODY_CHUNK = 0x62647461

TAG_FOR_QUASI_HTTP_BODY_CHUNK_EXT = 0x62657874

DEFAULT_MAX_LENGTH  = 134_217_728 # 128 MB

def encode_tag_and_length(tag: int, length: int):
    if tag <= 0:
        raise KabomuIOError(f"invalid tag: {tag}")
    if length < 0:
        raise KabomuIOError(f"invalid length: {length}")
    encoded_tag = misc_utils_internal.serialize_int32_be(tag)
    return encoded_tag + misc_utils_internal.serialize_int32_be(length)

def decode_tag(data, offset):
    tag = misc_utils_internal.deserialize_int32_be(data, offset)
    if tag <= 0:
        raise IllegalArgumentError(f"invalid tag: {tag}")
    return tag

def decode_length(data, offset):
    decoded_length = misc_utils_internal.deserialize_int32_be(data, offset)
    if decoded_length < 0:
        raise IllegalArgumentError(f"invalid length: {decode_length}")
    return decoded_length

def create_content_length_enforcing_stream(backing_stream: ReceiveStream,
                                           content_length: int):
    class ContentLengthEnforcingStream(ReceiveStream):
        def __init__(self):
            super().__init__()
            self._bytes_left_to_read = int(content_length)

        async def receive_some(self, max_bytes=None):
            if not max_bytes:
                max_bytes = 8_192
            bytes_to_read = min(self._bytes_left_to_read, int(max_bytes))
            next_chunk = b""
            if bytes_to_read:
                next_chunk = await io_utils_internal.receive_some(
                    backing_stream, bytes_to_read)
            next_chunk_len = len(next_chunk)
            self._bytes_left_to_read -= next_chunk_len
            end_of_read = not next_chunk_len
            if end_of_read and self._bytes_left_to_read > 0:
                raise KabomuIOError.create_end_of_read_error()
            return next_chunk

        async def aclose(self):
            pass
        
    return ContentLengthEnforcingStream()

def create_max_length_enforcing_stream(backing_stream: ReceiveStream,
                                       max_length=None):
    class MaxLengthEnforcingStream(ReceiveStream):
        def __init__(self):
            super().__init__()
            nonlocal max_length
            if not max_length:
                max_length = DEFAULT_MAX_LENGTH
            self._bytes_left_to_read = int(max_length) + 1 # check for excess read.

        async def receive_some(self, max_bytes=None):
            if not max_bytes:
                max_bytes = 8_192
            bytes_to_read = min(self._bytes_left_to_read, int(max_bytes))
            next_chunk = b""
            if bytes_to_read:
                next_chunk = await io_utils_internal.receive_some(
                    backing_stream, bytes_to_read)
            self._bytes_left_to_read -= len(next_chunk)
            if not self._bytes_left_to_read:
                raise KabomuIOError(f"stream size exceeds limit of {max_length} bytes")
            return next_chunk

        async def aclose(self):
            pass

    return MaxLengthEnforcingStream()

def create_tlv_encoding_writable_stream(backing_stream: SendStream,
                                        tag_to_use: int):
    #class BodyChunkEncodingStream(SendStream, HalfCloseableStream):
    class BodyChunkEncodingStream:
        def __init__(self):
            pass
        
        async def send_eof(self):
            await io_utils_internal.send_all(
                backing_stream, encode_tag_and_length(tag_to_use, 0)
            )
        
        async def send_all(self, data):
            data_len = len(data)
            if data_len:
                await io_utils_internal.send_all(
                    backing_stream,
                    encode_tag_and_length(tag_to_use, data_len))
                await io_utils_internal.send_all(
                    backing_stream, data)

        def wait_send_all_might_not_block(self):
            return backing_stream.wait_send_all_might_not_block()

        async def aclose(self):
            pass

    return BodyChunkEncodingStream()

def create_tlv_decoding_readable_stream(
        backing_stream: ReceiveStream,
        expected_tag: int, tag_to_ignore: int):
    class BodyChunkDecodingStream(ReceiveStream):
        def __init__(self) -> None:
            super().__init__()
            self._last_chunk_seen = False
            self._chunk_data_len_rem = 0

        async def receive_some(self, max_bytes=None):
            if not max_bytes:
                max_bytes = 8_192
            # once empty data chunk is seen,
            # return empty for all subsequent reads.
            if self._last_chunk_seen:
                return b""
            
            if not self._chunk_data_len_rem:
                self._chunk_data_len_rem = await self._fetch_next_tag_and_length()
                if not self._chunk_data_len_rem:
                    self._last_chunk_seen = True
                    return b""
                
            max_bytes = min(self._chunk_data_len_rem, int(max_bytes))
            next_chunk = await io_utils_internal.receive_some(
                backing_stream, max_bytes)
            next_chunk_len = len(next_chunk)
            if not next_chunk_len:
                raise KabomuIOError.create_end_of_read_error()
            self._chunk_data_len_rem -= next_chunk_len
            return next_chunk
        
        async def _fetch_next_tag_and_length(self):
            tag = await self._read_tag_only()
            if tag == tag_to_ignore:
                await self._read_away_tag_value()
                tag = await self._read_tag_only()
            
            if tag != expected_tag:
                raise KabomuIOError(
                    f"unexpected tag: expected {expected_tag} but found {tag}")
            return await self._read_length_only()
        
        async def _read_away_tag_value(self):
            length = await self._read_length_only()
            src = create_content_length_enforcing_stream(
                backing_stream, length)
            async for _ in src:
                pass

        async def _read_tag_only(self):
            encoded_tag = await io_utils_internal.read_bytes_fully(
                backing_stream, 4)
            tag = int.from_bytes(encoded_tag, signed=True)
            if tag <= 0:
                raise KabomuIOError(f"invalid tag: {tag}")
            return tag

        async def _read_length_only(self):
            encoded_length = await io_utils_internal.read_bytes_fully(
                backing_stream, 4)
            length = int.from_bytes(encoded_length, signed=True)
            if length < 0:
                raise KabomuIOError(f"invalid tag value length: {length}")
            return length

        async def aclose(self):
            pass

    return BodyChunkDecodingStream()



from kabomu.errors import IllegalArgumentError

def serialize_int32_be(v: int):
    return v.to_bytes(4, signed=True)

def deserialize_int32_be(src, offset: int):
    if not is_valid_byte_buffer_slice(src, offset, 4):
        raise IllegalArgumentError("invalid byte buffer slice")
    return int.from_bytes(src[offset:offset+4], signed=True)

def parse_int_48(input):
    n = int(input)
    if n < -140_737_488_355_328 or n > 140_737_488_355_327:
        raise IllegalArgumentError(f"invalid 48-bit integer: {n}")
    return n

def parse_int_32(input):
    n = int(input)
    if n < -2_147_483_648 or n > 2_147_483_647:
        raise IllegalArgumentError(f"invalid 32-bit integer: {n}")
    return n

def is_valid_byte_buffer_slice(data, offset: int, length: int):
    if data == None:
        return False
    if type(offset) != int:
        return False
    if type(length) != int:
        return False
    if offset < 0:
        offset += len(data)
    if offset < 0:
        return False
    if length < 0:
        return False
    if offset + length > len(data):
        return False
    return True

def string_to_bytes(s):
    return s.encode()

def bytes_to_string(raw_bytes):
    return raw_bytes.decode()


import sys

from kabomu.quasi_http_utils import _get_optional_attr
from kabomu import protocol_utils_internal
from kabomu.errors import MissingDependencyError,\
    QuasiHttpError,\
    QUASI_HTTP_ERROR_REASON_GENERAL

class StandardQuasiHttpServer:
    def __init__(self, transport=None, application=None):
        self.transport = transport
        self.application = application

    async def accept_connection(self, connection):
        if not connection:
            raise ValueError("connection argument is null")

        # access fields for use per processing call, in order to cooperate with
        # any implementation of field accessors which supports
        # concurrent modifications.
        transport = self.transport
        application = self.application
        if not transport:
            raise MissingDependencyError("server transport")
        if not application:
            raise MissingDependencyError("server application")

        try:
            async def proc():
                await _process_accept(
                    application, transport, connection)
            processed = await protocol_utils_internal.run_timeout_scheduler(
                connection, False, proc)
            if not processed:
                await _process_accept(application, transport,
                                      connection)
            await transport.release_connection(connection)
        except:
            #raise
            await _abort(transport, connection, True)
            ex = sys.exc_info()[1]
            if isinstance(ex, QuasiHttpError):
                raise
            abort_error = QuasiHttpError(
                QUASI_HTTP_ERROR_REASON_GENERAL,
                "encountered error during receive request processing")
            abort_error.__cause__ = ex
            raise abort_error

async def _process_accept(application, transport, connection):
    request = None
    if hasattr(transport, "request_deserializer"):
        request = await transport.request_deserializer(connection)
    if not request:
        request = await protocol_utils_internal.read_entity_from_transport(
            False, transport.get_readable_stream(connection),
            connection)

    response = await application.process_request(request)
    if not response:
        raise QuasiHttpError("no response")

    try:
        response_serialized = False
        if hasattr(transport, "response_serializer"):
            response_serialized = await transport.response_serializer(
                connection, response)
        if not response_serialized:
            await protocol_utils_internal.write_entity_to_transport(
                True, response, transport.get_writable_stream(connection),
                connection)
    finally:
        if hasattr(response, "release"):
            await response.release()

async def _abort(transport, connection, error_occured):
    if error_occured:
        try:
            # don't wait
            await transport.release_connection(
                connection) # swallow errors
        except:
            pass # ignore
    else:
        await transport.release_connection(
                connection)



from kabomu.errors import KabomuIOError

DEFAULT_READ_BUFFER_SIZE = 8192

async def receive_some(stream, max_bytes=None):
    src_is_file = hasattr(stream, "read")
    if src_is_file:
        if not max_bytes or max_bytes < 1:
            max_bytes = DEFAULT_READ_BUFFER_SIZE
        next_chunk = await stream.read(max_bytes)
    else:
        next_chunk = await stream.receive_some(max_bytes)
    assert next_chunk is not None
    return next_chunk

async def send_all(stream, data):
    dest_is_file = hasattr(stream, "write")
    if dest_is_file:
        next_chunk_len = len(data)
        bytes_written = 0
        while bytes_written < next_chunk_len:
            # cause error if write returns None
            if bytes_written == 0:
                bytes_written += int(await stream.write(
                    data))
            else:
                bytes_written += int(await stream.write(
                    data[bytes_written:]))
    else:
        await stream.send_all(data)

async def read_bytes_fully(stream, count: int):
    raw_data = bytearray()
    bytes_left = int(count)
    while bytes_left:
        next_chunk = await receive_some(stream, bytes_left)
        next_chunk_len = len(next_chunk)
        if not next_chunk_len:
            raise KabomuIOError.create_end_of_read_error()
        raw_data.extend(next_chunk)
        bytes_left -= next_chunk_len
    return raw_data


async def copy(src, dest):
    while True:
        next_chunk = await receive_some(src)
        next_chunk_len = len(next_chunk)
        if not next_chunk_len:
            break
        await send_all(dest, next_chunk)



import csv
import io

from trio.abc import ReceiveStream, SendStream

from kabomu.abstractions import DefaultQuasiHttpResponse,\
    DefaultQuasiHttpRequest
from kabomu.quasi_http_utils import _get_optional_attr
from kabomu.tlv import tlv_utils
from kabomu import misc_utils_internal, io_utils_internal, quasi_http_utils
from kabomu.errors import IllegalArgumentError, ExpectationViolationError,\
    MissingDependencyError,\
    QuasiHttpError,\
    QUASI_HTTP_ERROR_REASON_GENERAL,\
    QUASI_HTTP_ERROR_REASON_TIMEOUT,\
    QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,\
    QUASI_HTTP_ERROR_REASON_MESSAGE_LENGTH_LIMIT_EXCEEDED

async def run_timeout_scheduler(
        connection, for_client, proc):
    if not hasattr(connection, "schedule_timeout"):
        return
    result = await connection.schedule_timeout(proc)
    if result is None:
        return
    error = _get_optional_attr(result, "error")
    if error:
        raise error
    if _get_optional_attr(result, "timeout"):
        timeout_msg = "send timeout" if for_client else "receive timeout"
        raise QuasiHttpError(QUASI_HTTP_ERROR_REASON_TIMEOUT,
            timeout_msg)

    if for_client:
        response = _get_optional_attr(result, "response")
        if response is None:
            raise QuasiHttpError(QUASI_HTTP_ERROR_REASON_GENERAL,
                "no response from timeout scheduler")
        return response
    else:
        return True

def validate_http_header_section(is_response, csv_data):
    if not csv_data:
        raise ExpectationViolationError(
            "expected csv to contain at least the special header")
    special_header = csv_data[0]
    if len(special_header) != 4:
        raise ExpectationViolationError(
            "expected special header to have 4 values " +
            f"instead of {len(special_header)}")
    error_tag = "status" if is_response else "request"
    for i in range(len(special_header)):
        item = special_header[i]
        if not contains_only_printable_ascii_chars(
                item, is_response and i == 2):
            raise QuasiHttpError(
                QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
                    f"quasi http {error_tag} line " +
                    "field contains spaces, newlines or " +
                    "non-printable ASCII characters: " +
                    item)
    for row in csv_data[1:]:
        if len(row) < 2:
            raise ExpectationViolationError(
                "expected row to have at least 2 values " +
                f"instead of {len(row)}")
        header_name = row[0]
        if not contains_only_header_name_chars(header_name):
            raise QuasiHttpError(
                QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
                    "quasi http header name contains characters " +
                    "other than hyphen and English alphabets: " +
                    header_name)
        for header_value in row[1:]:
            if not contains_only_printable_ascii_chars(header_value, True):
                raise QuasiHttpError(
                    QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
                        "quasi http header value contains newlines or " +
                        "non-printable ASCII characters: " +
                        header_value)

def contains_only_header_name_chars(v):
    for c in v:
        c = ord(c)
        if c >= 48 and c < 58:
            # digits.
            pass
        elif c >= 65 and c < 91:
            # upper case
            pass
        elif c >= 97 and c < 123:
            # lower case
            pass
        elif c == 45:
            # hyphen
            pass
        else:
            return False
    return True

def contains_only_printable_ascii_chars(v, allow_space):
    for c in v:
        c = ord(c)
        if c < 32 or c > 126:
            return False
        if not allow_space and c == 32:
            return False
    return True

def encode_quasi_http_headers(is_response, req_or_status_line,
                              remaining_headers):
    if not req_or_status_line:
        raise IllegalArgumentError("req_or_status_line argument is null")
    csv_data = []
    special_header = []
    for v in req_or_status_line:
        special_header.append('' if v is None else str(v))
    csv_data.append(special_header)
    if remaining_headers:
        for header_name, header_value in remaining_headers.items():
            header_name = '' if header_name is None else str(header_name)
            if not header_name:
                raise QuasiHttpError(QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
                    "quasi http header name cannot be empty")
            # allow string values not inside an array.
            if type(header_value) == str:
                header_value = [header_value]
            if header_value is None or not len(header_value):
                continue
            header_row = [header_name]
            for v in header_value:
                v = "" if v is None else str(v)
                if not v:
                    raise QuasiHttpError(QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
                        "quasi http header value cannot be empty")
                header_row.append(v)
            csv_data.append(header_row)

    validate_http_header_section(is_response, csv_data)

    with io.StringIO(newline='') as serialized:
        for row in csv_data:
            add_comma = False
            for col in row:
                if add_comma:
                    serialized.write(',')
                serialized.write('"' + col.replace('"', '""') + '"')
                add_comma = True
            serialized.write('\n')
        return misc_utils_internal.string_to_bytes(
            serialized.getvalue())
    
def decode_quasi_http_headers(is_response, buffer,
                              headers_receiver):
    csv_data = []
    try:
        with io.StringIO(misc_utils_internal.bytes_to_string(
                buffer)) as s:
            deserializer = csv.reader(s,strict=True)
            for row in deserializer:
                csv_data.append(row)
    except Exception as ex:
        quasiHttpEx = QuasiHttpError(
            QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
            "invalid quasi http headers")
        quasiHttpEx.__cause__ = ex
        raise quasiHttpEx
    if not csv_data:
        raise QuasiHttpError(
            QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
            "invalid quasi http headers")
    
    error_tag = "status" if is_response else "request"
    special_header = csv_data[0]
    if len(special_header) < 4:
        raise QuasiHttpError(
            QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
            f"invalid quasi http {error_tag} line")
    for header_row in csv_data[1:]:
        if len(header_row) < 2:
            continue
        # merge headers with the same normalized name in different rows.
        header_name = header_row[0].lower()
        if header_name not in headers_receiver:
            headers_receiver[header_name] = []
        header_value = header_row[1:]
        headers_receiver[header_name].extend(header_value)
    return special_header

async def write_quasi_http_headers(
        is_response, dest: SendStream, req_or_status_line,
        remaining_headers, max_headers_size):
    encoded_headers = encode_quasi_http_headers(
        is_response, req_or_status_line, remaining_headers)
    if not max_headers_size or max_headers_size < 0:
        max_headers_size = quasi_http_utils.DEFAULT_MAX_HEADERS_SIZE

    # finally check that byte count of csv doesn't exceed limit.
    if len(encoded_headers) > max_headers_size:
        msg_suffix = f"{len(encoded_headers)} > {max_headers_size}"
        raise QuasiHttpError(QUASI_HTTP_ERROR_REASON_MESSAGE_LENGTH_LIMIT_EXCEEDED,
            f"quasi http headers exceed max size ({msg_suffix})")

    tag_and_len = tlv_utils.encode_tag_and_length(
        tlv_utils.TAG_FOR_QUASI_HTTP_HEADERS,
        len(encoded_headers))
    await dest.send_all(tag_and_len)
    await dest.send_all(encoded_headers)

async def read_quasi_http_headers(
        is_response, src: ReceiveStream,
        headers_receiver, max_headers_size):
    encoded_tag = await io_utils_internal.read_bytes_fully(src, 4)
    tag = tlv_utils.decode_tag(encoded_tag, 0)
    if tag != tlv_utils.TAG_FOR_QUASI_HTTP_HEADERS:
        raise QuasiHttpError(
            QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
            f"unexpected quasi http headers tag: ${tag}")

    if not max_headers_size or max_headers_size < 0:
        max_headers_size = quasi_http_utils.DEFAULT_MAX_HEADERS_SIZE
    encoded_len = await io_utils_internal.read_bytes_fully(src, 4)
    headers_size = tlv_utils.decode_length(encoded_len, 0)
    if headers_size > max_headers_size:
        msg_suffix = f"{headers_size} > {max_headers_size}"
        raise QuasiHttpError(QUASI_HTTP_ERROR_REASON_MESSAGE_LENGTH_LIMIT_EXCEEDED,
            f"quasi http headers exceed max size ({msg_suffix})")
    encoded_headers = await io_utils_internal.read_bytes_fully(
        src, headers_size)
    return decode_quasi_http_headers(is_response, encoded_headers,
        headers_receiver)

async def write_entity_to_transport(
        is_response, entity, writable_stream: SendStream,
        connection):
    if not writable_stream:
        raise MissingDependencyError("no writable stream found for transport")
    if is_response:
        response = entity
        headers = _get_optional_attr(response, "headers")
        body = _get_optional_attr(response, "body")
        content_length = _get_optional_attr(response, "content_length")
        if not content_length:
            content_length = 0
        status_code = _get_optional_attr(response, "status_code")
        if not status_code:
            status_code = 0
        req_or_status_line = []
        req_or_status_line.append(
            _get_optional_attr(response, "http_version"))
        req_or_status_line.append(status_code)
        req_or_status_line.append(
            _get_optional_attr(response, "http_status_message"))
        req_or_status_line.append(content_length)
    else:
        request = entity
        headers = _get_optional_attr(request, "headers")
        body = _get_optional_attr(request, "body")
        content_length = _get_optional_attr(request, "content_length")
        if not content_length:
            content_length = 0
        req_or_status_line = []
        req_or_status_line.append(
            _get_optional_attr(request, "http_method"))
        req_or_status_line.append(
            _get_optional_attr(request, "target"))
        req_or_status_line.append(
            _get_optional_attr(request, "http_version"))
        req_or_status_line.append(content_length)
    # treat content lengths totally separate from body
    # due to how HEAD method works.
    max_headers_size = _get_optional_attr(
        connection, "processing_options", "max_headers_size")
    await write_quasi_http_headers(is_response,
        writable_stream, req_or_status_line, headers,
        max_headers_size)
    if not body:
        # don't proceed, even if content length is not zero.
        return
    if content_length > 0:
        # don't enforce positive content lengths when writing out
        # quasi http bodies
        await io_utils_internal.copy(body, writable_stream)
    else:
        # proceed, even if content length is 0.
        body_writer = tlv_utils.create_tlv_encoding_writable_stream(
            writable_stream, tlv_utils.TAG_FOR_QUASI_HTTP_BODY_CHUNK)
        await io_utils_internal.copy(body, body_writer)
        await body_writer.send_eof()

async def read_entity_from_transport(
        is_response, readable_stream: ReceiveStream, connection):
    if not readable_stream:
        raise MissingDependencyError("no readable stream found for transport")
    headers_receiver = {}
    processing_options = _get_optional_attr(
        connection, "processing_options")
    max_headers_size = _get_optional_attr(
        processing_options, "max_headers_size")
    req_or_status_line = await read_quasi_http_headers(
        is_response, readable_stream, headers_receiver,
        max_headers_size)
    error_tag = 'response' if is_response else 'request'
    try:
        content_length = misc_utils_internal.parse_int_48(
            req_or_status_line[3])
    except Exception as ex:
        quasi_ex = QuasiHttpError(QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
            f"invalid quasi http {error_tag} content length")
        quasi_ex.__cause__ = ex
        raise quasi_ex
    body = None
    if content_length:
        if content_length > 0:
            body = tlv_utils.create_content_length_enforcing_stream(
                readable_stream, content_length)
        else:
            body = tlv_utils.create_tlv_decoding_readable_stream(
                readable_stream,
                tlv_utils.TAG_FOR_QUASI_HTTP_BODY_CHUNK,
                tlv_utils.TAG_FOR_QUASI_HTTP_BODY_CHUNK_EXT)
    if is_response:
        response = DefaultQuasiHttpResponse()
        response.http_version = req_or_status_line[0]
        try:
            response.status_code = misc_utils_internal.parse_int_32(
                req_or_status_line[1])
        except Exception as ex:
            quasi_ex = QuasiHttpError(QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION,
                "invalid quasi http response status code")
            quasi_ex.__cause__ = ex
            raise quasi_ex
        response.http_status_message = req_or_status_line[2]
        response.content_length = content_length
        response.headers = headers_receiver
        if body:
            body_size_limit = None
            if processing_options:
                body_size_limit = _get_optional_attr(
                    processing_options, "max_response_body_size")
                if not body_size_limit or body_size_limit > 0:
                    body = tlv_utils.create_max_length_enforcing_stream(
                        body, body_size_limit)
        response.body = body
        return response
    else:
        request = DefaultQuasiHttpRequest()
        request.environment = _get_optional_attr(
            connection, "environment")
        request.http_method = req_or_status_line[0]
        request.target = req_or_status_line[1]
        request.http_version = req_or_status_line[2]
        request.content_length = content_length
        request.headers = headers_receiver
        request.body = body
        return request



class KabomuError(Exception):
    def __init__(self, *args: object):
        super().__init__(*args)

class KabomuIOError(KabomuError):
    def __init__(self, *args: object):
        super().__init__(*args)

    @staticmethod
    def create_end_of_read_error():
        return KabomuIOError("unexpected end of read")
    
class ExpectationViolationError(KabomuError):
    def __init__(self, *args: object):
        super().__init__(*args)

class MissingDependencyError(KabomuError):
    def __init__(self, *args: object):
        super().__init__(*args)

class IllegalArgumentError(KabomuError):
    def __init__(self, *args: object):
        super().__init__(*args)

QUASI_HTTP_ERROR_REASON_GENERAL = 1
QUASI_HTTP_ERROR_REASON_TIMEOUT = 2
QUASI_HTTP_ERROR_REASON_PROTOCOL_VIOLATION = 3
QUASI_HTTP_ERROR_REASON_MESSAGE_LENGTH_LIMIT_EXCEEDED = 4

class QuasiHttpError(KabomuError):
    def __init__(self, reason_code, *args: object):
        super().__init__(*args)
        if not reason_code or (reason_code > 4 and reason_code < 10):
            raise IllegalArgumentError(f"cannot use reserved reason code: {reason_code}")
        self.reason_code = reason_code 